#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>

#include "lisp.h"

/** CONSTRUCTORS **/
Pointer number(Number num) {
  Pointer ptr = memory_new(NUMBER, sizeof(Number));
  NUMBER(ptr) = num;
  return ptr;
}

Pointer native_func(NativeFunc func) {
  Pointer ptr = memory_new(NATIVE_FUNC, sizeof(NativeFunc));
  NATIVE_FUNC(ptr) = func;
  return ptr;
}

Pointer special_form(SpecialForm func) {
  Pointer ptr = memory_new(SPECIAL_FORM, sizeof(SpecialForm));
  SPECIAL_FORM(ptr) = func;
  return ptr;
}

Pointer func(Pointer code, Pointer env) {
  Pointer fn = memory_new(FUNC, sizeof(Func));
  FUNC(fn).code = code;
  FUNC(fn).env = env;
  return fn;
}

Pointer macro(Pointer code, Pointer env) {
  Pointer fn = memory_new(MACRO, sizeof(Func));
  FUNC(fn).code = code;
  FUNC(fn).env = env;
  return fn;
}

/** FUNCTIONS **/

Pointer add_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer result = number(0);
  REDUCE(args, result; NUMBER(result) += NUMBER(CAR(args)), result);
  return result;
}

Pointer sub_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer first = CAR(args);
  Pointer rest = CDR(args);
  if (first == NIL) return UNDEFINED; // Empty args
  if (rest == NIL) return number(-NUMBER(first));
  Pointer result = number(NUMBER(first));
  REDUCE(rest, result; NUMBER(result) -= NUMBER(CAR(rest)), result);
  return result;
}

Pointer mul_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer result = number(1);
  REDUCE(args, result; NUMBER(result) *= NUMBER(CAR(args)), result);
  return result;
}

Pointer div_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer first = CAR(args);
  Pointer rest = CDR(args);
  if (first == NIL) return UNDEFINED; // Empty args
  if (rest == NIL) return number(1 / NUMBER(first));
  Pointer result = number(NUMBER(first));
  REDUCE(rest, result; NUMBER(result) /= NUMBER(CAR(rest)), result);
  return result;
}

Pointer pow_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer a = CAR(args);
  Pointer b = CAR(CDR(args));
  if (a == NIL || b == NIL) return UNDEFINED; // Arguments missing.
  return number(pow(NUMBER(a), NUMBER(b)));
}

Pointer sqrt_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer a = CAR(args);
  if (a == NIL) return UNDEFINED; // Arguments missing.
  return number(sqrt(NUMBER(a)));
}

Pointer logand_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer result = number(-1l);
  REDUCE(args,
	 result; NUMBER(result) = (long) NUMBER(result) & (long) NUMBER(CAR(args)),
	 result);
  return result;
}

Pointer logor_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer result = number(0);
  REDUCE(args,
	 result; NUMBER(result) = (long) NUMBER(result) | (long) NUMBER(CAR(args)),
	 result);
  return result;
}

Pointer logxor_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer result = number(0);
  REDUCE(args,
	 result; NUMBER(result) = (long) NUMBER(result) ^ (long) NUMBER(CAR(args)),
	 result);
  return result;
}

Pointer lognot_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer a = CAR(args);
  if (a == NIL) return UNDEFINED; // Arguments missing.
  return number(~(long) NUMBER(a));
}

Pointer if_fn(Pointer args, Pointer env) {
  if (eval_fn(args, env) != NIL) return eval_fn(CDR(args), env);
  return eval_fn(CDR(CDR(args)), env);
}

Pointer let_fn(Pointer args, Pointer env) {
  Pointer alist = CAR(args);
  Pointer body = CDR(args);
  Pointer tbl = NIL;
  Pointer keypair;
  Pointer key;
  Pointer value;

  if (alist != NIL) {
    tbl = table(2);
    env = cons(tbl, env);
    do {
      keypair = CAR(alist);
      key = CAR(keypair);
      value = eval_fn(CDR(keypair), env);
      table_set(tbl, key, value);
      alist = CDR(alist);
    } while(alist != NIL);
  }

  while(body != NIL) {
    value = eval_fn(body, env);
    body = CDR(body);
  } 
  return value;
}

Pointer quote_fn(Pointer args, Pointer env) {
  (void) env; return CAR(args);
}

Pointer and_fn(Pointer args, Pointer env) {
  Pointer value = NIL;
  while(args != NIL) {
    value = eval_fn(args, env);
    if (value == NIL) return NIL;
    args = CDR(args);
  }
  return value;
}

Pointer or_fn(Pointer args, Pointer env) {
  Pointer value = NIL;
  while(args != NIL) {
    value = eval_fn(args, env);
    if (value != NIL) return value;
    args = CDR(args);
  } 
  return NIL;
}

Pointer not_fn(Pointer args, Pointer env) {
  (void) env;
  return CAR(args) == NIL ? T : NIL;
}

Pointer def_fn(Pointer args, Pointer env) {
  Pointer key = CAR(args);
  Pointer value = eval_fn(CDR(args), env);
  environment_set(NIL, key, value);
  return key;
}

Pointer set_fn(Pointer args, Pointer env) {
  Pointer key = CAR(args);
  Pointer value = eval_fn(CDR(args), env);

  environment_set(env, key, value);
  return value;
}

Pointer fn_fn(Pointer args, Pointer env) {
  return func(args, env);
}

Pointer defmacro_fn(Pointer args, Pointer env) {
  return environment_set(NIL, CAR(args), macro(CDR(args), env));
}

Pointer exit_fn(Pointer args, Pointer env) {
  (void) args; (void) env;
  exit(1);
}

Pointer eq_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer a = CAR(args);
  Pointer b = CAR(CDR(args));
  Type type = TYPE(a);

  if (type != TYPE(b)) {
    return NIL;
  }

  switch (type) {
  case NUMBER: 
    return NUMBER(a) == NUMBER(b) ? T : NIL;
  case STRING:
    return strcmp(STRING(a).data, STRING(b).data) == 0 ? T : NIL;
  default:
    return a == b ? T : NIL;
  }
}

Pointer lt_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer a = CAR(args);
  Pointer b = CAR(CDR(args));
  if (TYPE(a) != NUMBER || TYPE(b) != NUMBER) {
    return UNDEFINED;
  }
  return NUMBER(a) < NUMBER(b) ? T : NIL;
}

Pointer gt_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer a = CAR(args);
  Pointer b = CAR(CDR(args));
  if (TYPE(a) != NUMBER || TYPE(b) != NUMBER) {
    return UNDEFINED;
  }
  return NUMBER(a) > NUMBER(b) ? T : NIL;
}

Pointer le_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer a = CAR(args);
  Pointer b = CAR(CDR(args));
  if (TYPE(a) != NUMBER || TYPE(b) != NUMBER) {
    return UNDEFINED;
  }
  return NUMBER(a) <= NUMBER(b) ? T : NIL;
}

Pointer ge_fn(Pointer args, Pointer env) {
  (void) env;
  Pointer a = CAR(args);
  Pointer b = CAR(CDR(args));
  if (TYPE(a) != NUMBER || TYPE(b) != NUMBER) {
    return UNDEFINED;
  }
  return NUMBER(a) >= NUMBER(b) ? T : NIL;
}

#define SET_FUNC(s, fn) environment_set(NIL, symbol1(s), native_func(fn))
#define SET_FORM(s, fn) environment_set(NIL, symbol1(s), special_form(fn))

void init(void) {
  memory_init(16);
  symbol_init();
  reader_init();
  environment_init();
  environment_set(NIL, STANDARD_INPUT, stream(stdin));
  environment_set(NIL, STANDARD_OUTPUT, stream(stdout));
  SET_FUNC("eval", eval_fn);
  SET_FUNC("cons", cons_fn);
  SET_FUNC("car", car_fn);
  SET_FUNC("cdr", cdr_fn);
  SET_FUNC("reduce", reduce_fn);
  SET_FUNC("list", list_fn);
  SET_FUNC("+", add_fn);
  SET_FUNC("-", sub_fn);
  SET_FUNC("*", mul_fn);
  SET_FUNC("/", div_fn);
  SET_FUNC("pow", pow_fn);
  SET_FUNC("sqrt", sqrt_fn);
  SET_FUNC("&", logand_fn);
  SET_FUNC("|", logor_fn);
  SET_FUNC("^", logxor_fn);
  SET_FUNC("~", lognot_fn);
  SET_FUNC("peek-char", peek_char_fn);
  SET_FUNC("read-char", read_char_fn);
  SET_FUNC("read", read_fn);
  SET_FUNC("set-reader-macro", set_reader_macro_fn);
  SET_FUNC("exit", exit_fn);
  SET_FUNC("=", eq_fn);
  SET_FUNC("<", lt_fn);
  SET_FUNC(">", gt_fn);
  SET_FUNC("<=", le_fn);
  SET_FUNC(">=", ge_fn);
  SET_FUNC("print", print_fn);

  SET_FORM("if", if_fn);
  SET_FORM("let", let_fn);
  SET_FORM("quote", quote_fn);
  SET_FORM("and", and_fn);
  SET_FORM("or", or_fn);
  SET_FORM("def", def_fn);
  SET_FORM("set", set_fn);
  SET_FORM("fn", fn_fn);
  SET_FORM("defmacro", defmacro_fn);
}

/*
(set-reader-macro \' (fn (s c) (list quote (read s))))
(defmacro defn (name args &body body) (list def name (cons fn (cons args body))))
 */

void repl(void) {
  while(true) {
    printf(">>> ");
    Pointer readed = read_fn(NIL, NIL);
    char c;
    while((c = getc(stdin)) && isspace(c) && c != '\n');
    if (c != '\n') ungetc(c, stdin);
    print(eval(readed, NIL), stdout);

    printf("\n");
  }
}
